#!/usr/bin/env python
# -*- coding: utf-8 -*-
from hashlib import sha256

import logbook
from base58 import b58encode

logger = logbook.Logger(__name__)


class AddressProvider(object):
    """
    AddressProvider is an abstraction allowing any string to be converted into an "Address".
    Address should have hash like properties of providing deterministic output for a given input and being
    non-reversable.
    """
    def get_address(self):
        raise NotImplementedError()


class CreditsAddressProvider(AddressProvider):
    def __init__(self, value):
        """
        An AddressProvider using an in-house algorithm based off double sha256 and base58 encoding.

        :type input_str: str
        """
        self.value = value

    def get_address(self) -> str:
        """
        Generate an address from self.input_str. This is typically performed using a verifying key.

        :rtype: str
        """
        digest = b"\x00" + sha256(self.value.encode()).digest()[:20]
        checksum = sha256(digest).digest()[:4]
        return b58encode(digest + checksum)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logbook

from credits.address import CreditsAddressProvider
from credits.interface import Applicable
from credits.interface import Marshallable
from credits.interface import Signable


class Proof(Marshallable, Applicable, Signable):
    def __init__(self):
        self.logger = logbook.Logger(__name__)


class SingleKeyProof(Proof):
    fqdn = 'works.credits.core.SingleKeyProof'

    STATE_NONCE = "works.credits.core.IntegerNonce"

    def __init__(self, address, nonce, challenge, verifying_key=None, signature=None):
        """
        Sign a challenge. This signature can be used as a proof.

        :type registry: credits.registry.ComponentRegistry
        :type address: str
        :type nonce: int
        :type challenge: str
        :type verifying_key: credits.key.VerifyingKey
        :type signature: str
        """
        super(SingleKeyProof, self).__init__()

        self.address = address
        self.nonce = nonce
        self.challenge = challenge

        self.verifying_key = verifying_key
        self.signature = signature

    def verify(self, state):
        """
        Verify this proof has been signed and that it's signature/verifying_key/challenge is valid against state.

        :returns: result, error
        """
        if (self.signature is None) or (self.verifying_key is None):
            error = "Proof has not been signed."
            self.logger.error(error)
            return None, Exception(error)

        address = CreditsAddressProvider(self.verifying_key.to_string()).get_address()

        if address != self.address:
            error = "Proof for address %s was signed with %s" % (self.address, address)
            self.logger.error("Proof for address %s was signed with %s" % (self.address, address))
            return None, Exception(error)

        if not self.verifying_key.verify(bytes.fromhex(self.challenge), bytes.fromhex(self.signature)):
            error = "SingleKeyProof failed a signature check against %s" % address
            self.logger.error(error)
            return None, Exception(error)

        known_nonce = state[self.STATE_NONCE][address]
        if self.nonce < known_nonce:
            error = "SingleKeyProof nonce (%s) is less than current nonce (%s) for %s" % (
                self.nonce,
                known_nonce,
                address,
            )
            self.logger.error(error)
            return None, Exception(error)

        return state, None

    def apply(self, state):
        """
        Increment nonce of the associated address. this operation should assume the address potentially doesn't exist so
        this operation should be performed safely.
        """
        address = CreditsAddressProvider(self.verifying_key.to_string()).get_address()
        nonces = state[self.STATE_NONCE]

        if self.nonce != nonces.get(address, 0):
            error = "SingleKeyProof nonce (%s) is not equal to nonce (%s) for %s" % (
                self.nonce,
                nonces.get(address, 0),
                address,
            )
            self.logger.error(error)
            return None, Exception(error)

        nonces[address] = nonces.get(address, 0) + 1

        return state, None

    def sign(self, signing_key):
        """
        Sign this proof.

        :type signing_key: credits.key.SigningKey
        :rtype: credits.proof.SingleKeyProof
        """
        verifying_key = signing_key.get_verifying_key()
        signature = signing_key.sign(bytes.fromhex(self.challenge)).hex()

        return SingleKeyProof(
            address=self.address,
            nonce=self.nonce,
            challenge=self.challenge,
            verifying_key=verifying_key,
            signature=signature,
        )

    def marshall(self):
        verifying_key = None
        if self.verifying_key is not None:
            verifying_key = self.verifying_key.marshall()

        return {
            "fqdn": self.fqdn,
            "address": self.address,
            "nonce": self.nonce,
            "challenge": self.challenge,
            "verifying_key": verifying_key,
            "signature": self.signature,
        }

    @classmethod
    def unmarshall(cls, registry, payload):
        return cls(
            address=payload["address"],
            nonce=payload["nonce"],
            challenge=payload["challenge"],
            verifying_key=registry.unmarshall(payload["verifying_key"]),
            signature=payload["signature"],
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from collections import UserDict


class DummyRegistry(UserDict):
    def register(self, marshallable):
        self[marshallable.fqdn] = marshallable

    def unmarshall(self, payload):
        return self[payload["fqdn"]].unmarshall(self, payload)


def check_transform(cls, dependencies, kwargs, state, state_expected):
    """
    Test a Tranform is implemented correctly.

    :param cls: Your Transform class.
    :param dependencies: A list of Marshallable objects. If you've not used any in your transform, pass an empty list.
    :param kwargs: A list of keyword arguments to be passed to the constructor of your Transform.
    :param state: A mocked up state for checking verify/apply logic.
    :param state: A mocked up resultant state for checking transform.apply applied correctly.

    :type cls: credits.transform.Transform
    :type dependencies: [credits.interface.Marshallable]
    :type kwargs: dict
    :type state: dict
    :type state_expected: dict
    """
    from copy import deepcopy
    from credits.hash import SHA256HashProvider

    hash_provider = SHA256HashProvider()

    # Setup registry
    registry = DummyRegistry()
    for dependency in dependencies:
        registry.register(dependency)

    # Check registration
    registry.register(cls)

    # Construct the Transform
    transform = cls(**kwargs)

    # Get a digest.
    digest = transform.hash(hash_provider)

    # Check this has a UNIQUE fqdn.
    fqdn = transform.fqdn

    # Marshall this object into a primative map
    payload = transform.marshall()

    # Make sure the fqdn is in the payload.
    assert payload["fqdn"] == fqdn

    # Get the transform back from it's marshalled output.
    transform = registry.unmarshall(payload)

    # Check that it still digests like it did before.
    assert transform.hash(hash_provider) == digest

    # Verify against some demo state that it can apply either NOW or in the FUTURE
    result, error = transform.verify(deepcopy(state))
    if error is not None:
        raise error

    assert result is None  # All good!

    # Apply against state, check application is as expected.
    result, error = transform.apply(deepcopy(state))
    if error is not None:
        raise error

    assert result == state_expected  # All good!



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logbook

from credits.digest import digest
from credits.interface import Applicable
from credits.interface import Hashable
from credits.interface import HashProvider
from credits.interface import Marshallable


class Transaction(Marshallable, Applicable, Hashable):
    fqdn = "works.credits.core.Transaction"

    def __init__(self, transform, proofs):
        """
        :type transform: credits.transform.Transform
        :type proofs: {str: credits.proof.Proof}
        :type digest: str
        """
        self.logger = logbook.Logger(__name__)

        self.transform = transform
        self.proofs = proofs

    # interface.Applicable
    def verify(self, state):
        """
        Perform validation of transform and proofs against a given state. Raise ValidationException if not valid.

        :param state: The state to be validated.
        :type state: credits.state.State
        """
        if len(self.proofs) == 0:
            error = "Invalid Transaction: tx contains no proofs to validate against."
            self.logger.error(error)
            return None, Exception(error)

        # Make sure that the required auths are a subset of all proofs in this transaction.
        if not set(self.transform.required_authorizations()).issubset(set(self.proofs)):
            error = "Required Authorizations for this Transaction are not met."
            self.logger.error(error)
            return None, Exception(error)

        for address, proof in self.proofs.items():
            # Check the proof is valid.
            result, error = proof.verify(state)
            if error is not None:
                return None, error

            # Now check k == v for proofs by asking for the address the proof is associated with.
            if address != proof.address:
                error = "A proof for address %s does not contain an associated signature." % (address,)
                self.logger.error(error)
                return None, Exception(error)

        result, error = self.transform.verify(state)
        if error is not None:
            return None, error

        return state, None

    # interface.Applicable
    def apply(self, state):
        """
        Applies the transform, then the proofs. If all things apply correctly return the new state.

        :param state: The state to be modified.
        :type state: credits.state.State

        :rtype: credits.state.State
        """
        # Apply Proofs
        for proof in self.proofs.values():
            result, error = proof.apply(state)

            if error is not None:
                return None, error

        # Apply Transform
        result, error = self.transform.apply(state)
        if error is not None:
            return None, error

        return state, None

    def hash(self, hash_provider: HashProvider) -> str:
        items = [
            self.transform.get_challenge(hash_provider),
        ]

        for address, proof in self.proofs.items():
            items.append(address)
            items.append(proof.nonce)

        return hash_provider.hash(digest(items)).hex()

    # interface.Marshallable
    def marshall(self):
        proofs = {}
        for address, proof in self.proofs.items():
            proofs[address] = proof.marshall()

        return {
            "fqdn": self.fqdn,
            "transform": self.transform.marshall(),
            "proofs": dict((address, proof.marshall()) for address, proof in self.proofs.items()),
        }

    # interface.Marshallable
    @classmethod
    def unmarshall(cls, registry, payload):
        transform = registry.unmarshall(payload["transform"])
        proofs = dict((address, registry.unmarshall(proof)) for address, proof in payload["proofs"].items())

        return cls(
            transform=transform,
            proofs=proofs,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import hashlib

from credits.interface import HashProvider


class SHA256HashProvider(HashProvider):
    fqdn = "works.credits.core.SHA256HashProvider"

    def hash(self, i: bytes) -> bytes:
        """
        Return the SHA256 hash of the input i.

        :type i: bytes
        :rtype: bytes
        """
        return hashlib.sha256(i).digest()

    def marshall(self) -> dict:
        return {
            "fqdn": self.fqdn,
        }

    @classmethod
    def unmarshall(cls, registry, payload):
        return cls()


class SHA512HashProvider(HashProvider):
    fqdn = "works.credits.core.SHA512HashProvider"

    def hash(self, i: bytes) -> bytes:
        """
        Return the SHA512 hash of the input i.

        :type i: bytes
        :rtype: str
        """
        return hashlib.sha512(i).digest()

    def marshall(self) -> dict:
        return {
            "fqdn": self.fqdn,
        }

    @classmethod
    def unmarshall(cls, registry, payload):
        return cls()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logbook

from credits.digest import digest
from credits.interface import Hashable
from credits.interface import Marshallable


class Transform(Marshallable, Hashable):
    def __init__(self):
        self.logger = logbook.Logger(__name__)

    def verify(self, state):
        """
        Verify that this transform is still applicable in the future. Raise an exception if it can't verify.

        :param state: The state to verify against.
        :type: credits.state.State
        :return: bool
        """
        raise NotImplementedError()

    def apply(self, state):
        """
        Perform last minute checks that the transform can specifically apply _now_. Then apply. Return the new state.

        :param state: The state to apply against.
        :type: credits.state.State
        :return: credits.state.State
        """
        raise NotImplementedError()

    def required_authorizations(self):
        """
        Return a list of address's that must provide a proof for the parent transaction apply correctly.

        :return:
        """
        raise NotImplementedError()

    def hash(self, hash_provider):
        return hash_provider.hash(digest(self.marshall()))

    def get_challenge(self, hash_provider):
        return self.hash(hash_provider)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
from typing import Type
from typing import TypeVar

import ed25519
import logbook

from credits.interface import Marshallable

SigningKeyImpl = TypeVar("SigningKeyImpl", bound="SigningKey")
VerifyingKeyImpl = TypeVar("VerifyingKeyImpl", bound="VerifyingKey")


class SigningKey(Marshallable):
    fqdn = "NOTSET"

    def __init__(self):
        self.logger = logbook.Logger(__name__)

    @classmethod
    def new(cls: Type[SigningKeyImpl]) -> SigningKeyImpl:
        raise NotImplementedError()

    @classmethod
    def from_string(cls: Type[SigningKeyImpl], sk_s: str) -> SigningKeyImpl:
        raise NotImplementedError()

    def to_string(self) -> str:
        raise NotImplementedError()

    def get_verifying_key(self) -> VerifyingKeyImpl:
        raise NotImplementedError()

    def sign(self, data: bytes) -> bytes:
        raise NotImplementedError()

    def __str__(self):
        return self.to_string()


class VerifyingKey(Marshallable):
    fqdn = "NOTSET"

    def __init__(self):
        self.logger = logbook.Logger(__name__)

    @classmethod
    def from_string(cls: Type[VerifyingKeyImpl], sk_s: str) -> VerifyingKeyImpl:
        raise NotImplementedError()

    def to_string(self) -> str:
        raise NotImplementedError()

    def verify(self, data: bytes, signature: bytes) -> bool:
        raise NotImplementedError()

    def __str__(self):
        return self.to_string()


class ED25519VerifyingKey(VerifyingKey):
    fqdn = "works.credits.core.ED25519VerifyingKey"

    def __init__(self, vk):
        super(ED25519VerifyingKey, self).__init__()
        self.vk = vk

    @classmethod
    def unmarshall(cls, registry, payload):
        return cls.from_string(payload["vks"])

    def marshall(self):
        return {
            "fqdn": self.fqdn,
            "vks": self.to_string(),
        }

    @classmethod
    def from_string(cls, vks: str) -> "ED25519VerifyingKey":
        """
        Take the output from ED25519VerifyingKey.to_string and convert it back into a VerifyingKey.
        """
        vk = ed25519.VerifyingKey(bytes.fromhex(vks))
        return cls(vk)

    def to_string(self) -> str:
        """
        Generate a string representation of this verifying key, useful for generating addresses using an AddressProvider
        """
        # It gives back the hex. As bytes? Decode it here to get a str of the hex.
        return self.vk.to_ascii(encoding="hex").decode()

    def verify(self, msg: bytes, signature: bytes) -> bool:
        """
        Verify data with a provided signature.
        """
        try:
            # This library does arguments round the other way.
            self.vk.verify(signature, msg)
            return True

        except ed25519.BadSignatureError:
            return False


class ED25519SigningKey(SigningKey):
    fqdn = "works.credits.core.ED25519SigningKey"

    def __init__(self, sk):
        super(ED25519SigningKey, self).__init__()
        self.sk = sk

    @classmethod
    def unmarshall(cls, registry, payload):
        return cls.from_string(payload["sks"])

    def marshall(self):
        return {
            "fqdn": self.fqdn,
            "sks": self.to_string(),
        }

    @classmethod
    def new(cls) -> "ED25519SigningKey":
        """
        Generate a new ED25519SigningKey.
        """
        sk, _ = ed25519.create_keypair()
        return cls(sk)

    @classmethod
    def from_string(cls, sks):
        """
        Take the output from ED25519SigningKey.to_string and convert it back into a SigningKey.
        """
        sk = ed25519.SigningKey(codecs.decode(sks, "hex"))
        return cls(sk)

    def to_string(self):
        return self.sk.to_ascii(encoding="hex").decode()

    def get_verifying_key(self):
        """
        Generate the complementary ED25519VerifyingKey from this ED25519SigningKey.
        """
        return ED25519VerifyingKey(self.sk.get_verifying_key())

    def sign(self, data: bytes) -> bytes:
        return self.sk.sign(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import abc
from typing import TypeVar

HashableImpl = TypeVar('HashableImpl', bound='Hashable')
SignableImpl = TypeVar('SignableImpl', bound='Signable')


class Marshallable(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def marshall(self):
        """
        Convert this class into a primitive map.
        """

    @classmethod
    @abc.abstractmethod
    def unmarshall(cls, registry, payload):
        """
        Take a primitive map and restore it into an instance.
        """


class HashProvider(Marshallable):
    __metaclass__ = abc.ABCMeta

    fqdn = "NOTSET"

    @abc.abstractmethod
    def hash(self, i: bytes) -> bytes:
        """
        Given bytes input of 'i', return a cryptographic hash of it's contents.
        """


class Applicable(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def verify(self, state):
        """
        Verify this object against a given state. Throw exception if not valid.

        :type state: credits.state.State
        """

    @abc.abstractmethod
    def apply(self, state):
        """
        Apply this object to the state, mutate the state and return it.

        :type state: credits.state.State
        :rtype state: credits.state.State
        """


class Signable(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def sign(self, signing_key: 'SigningKey') -> SignableImpl:
        """
        Sign this object based on it's contents. Then return a new instance of the object with it's signature and
        verifying_key.

        :type signing_key: credits.key.SigningKey

        :returns: A signature of the contents of this object.
        :rtype: self
        """


class Hashable(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def hash(self, hash_provider: HashProvider) -> HashableImpl:
        """
        Hash this object based on it's contents. Then return a new instance of the object with it's hash.
        :type hash_provider: credits.hash.HashProvider

        :returns: A signature of the contents of this object.
        :rtype: self
        """


class Transactional(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def rollback(self):
        """
        Revert the unconfirmed changes to this Transactional object.
        """

    @abc.abstractmethod
    def commit(self):
        """
        Commit the unconfirmed changes to this Transactional object.
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import functools
from operator import itemgetter

NONE = b"\x00"
BOOL = b"\x01"
INT = b"\x02"
UTF8 = b"\x03"
BYTES = b"\x04"
LIST = b"\x05"
MAP = b"\x06"


@functools.singledispatch
def digest(obj):
    raise TypeError("Unknown type: %s." % (type(obj, )))


@digest.register(bytes)  # noqa
def _(obj: bytes):
    return BYTES + len(obj).to_bytes(4, "big") + obj


@digest.register(str)  # noqa
def _(obj: str):
    return UTF8 + len(obj).to_bytes(4, "big") + obj.encode("utf8")


@digest.register(int)  # noqa
def _(obj: int):
    return INT + obj.to_bytes(4, "big")


@digest.register(dict)  # noqa
def _(obj: dict):
    pairs = (digest(k) + digest(v) for (k, v) in sorted(obj.items(), key=itemgetter(0)))
    return MAP + len(obj).to_bytes(4, "big") + b"".join(pairs)


@digest.register(tuple)  # noqa
@digest.register(list)  # noqa
def _(container: list):
    return LIST + len(container).to_bytes(4, "big") + b"".join(digest(value) for value in container)


@digest.register(type(None))  # noqa
def _(value: type(None)):
    return NONE



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools

version = "3.0.0"

packages = [
    "base58==0.2.2",
    "logbook==0.12.5",
    "ed25519",  # specifically unversioned
]

test_packages = [
    "pytest", "mypy-lang", "pytest-mypy"
]

setuptools.setup(
    name="credits.common",
    version=version,
    author="Credits Developers",
    author_email="admin@credits.vision",
    description="Credits blockchain framework common client library",
    install_requires=packages,
    namespace_packages=['credits'],
    packages=setuptools.find_packages(),
    url="https://github.com/CryptoCredits/credits-common",
    download_url="https://github.com/CryptoCredits/credits-common/tarball/%s" % version,
    extras_require={
        'test': test_packages,
    },
)
